"""
Reader objects that handle different types of data
"""
import abc
import collections.abc
import gzip
import logging
import os
import sys
import typing as ty

import pysam

from . import (
    exceptions,
    parsers,
)


logger = logging.getLogger(__name__)


class BaseReader(abc.ABC):
    """Implements common base functionality for reading and filtering GWAS results"""
    def __init__(self,
                 source: ty.Any,
                 parser: ty.Union[ty.Callable[[str], object], None] = parsers.TupleLineParser(),
                 skip_rows: int = 0,
                 skip_errors: bool = False,
                 max_errors: int = 100,
                 **kwargs):
        """
        :param source: How to find the source data (implementation detail of the subclass)
        :param parser: A callable that parses a line of text. If `None`, the reader will not attempt to parse.
        :param skip_rows: How many header rows to skip when iterating over the entire file
        :param skip_errors: Whether to continue reading if a few bad lines are encountered
        :param max_errors: How many bad lines to accept (in skip_errors mode) before giving up
        :param kwargs:
        """
        self._source = source

        # Filters, lookups, and mutations that should operate on each row
        # If no parser is provided, just splits the row into a tuple based on the specified delimiter
        self._parser = parser
        self._filters = []  # type: list  # Should we return this row from iteration?
        self._lookups = []  # type: list  # Find the value of a specified field given (parsed) variant info
        self._transforms = []  # type: list  # Modify the (parsed) variant info using a custom function.

        # If using "skip error" mode, store a record of which lines had a problem (up to a point)
        self._skip_errors = skip_errors
        self._max_errors = max_errors
        self.errors = []  # type: list

        # The user must specify how many header rows to skip
        self._skip_rows = skip_rows

    @abc.abstractmethod
    def _create_iterator(self) -> ty.Iterator[str]:
        """Create an iterator that can be used over all possible data. Eg, open a file or go through a list"""
        raise NotImplementedError

    ######
    # Internal helper methods; should not need to override below this line

    def _make_generator(self, iterator: ty.Iterator[str]) -> ty.Iterator:
        """
        Process the output of an iterable before returning it. In the usual case where a parser is provided,
            this will parse the row, apply filter criteria, and exclude any rows with errors (though errors will be
            available for inspection later)
        """
        use_filters = len(self._filters)
        use_lookups = len(self._lookups)
        use_transforms = len(self._transforms)
        for i, row in enumerate(iterator):
            if not row:
                # Skip blank lines (eg at end of file)
                continue

            if not self._parser:
                # There is a "parser=None" option, to return raw lines of text. This is useful for, eg, format sniffers.
                yield row
            else:
                try:
                    parsed = self._parser(row)
                except exceptions.LineParseException as e:
                    if not self._skip_errors:
                        raise e
                    self.errors.append((i + self._skip_rows + 1, str(e), row))  # (human_line, message, raw_input)
                    if len(self.errors) >= self._max_errors:
                        raise exceptions.TooManyBadLinesException(error_list=self.errors)
                    continue

                if use_lookups:
                    for field_name, func in self._lookups:
                        setattr(parsed, field_name, func(parsed))

                if use_transforms:
                    for func in self._transforms:
                        parsed = func(parsed)

                if use_filters and not all(test_func(parsed) for test_func in self._filters):
                    continue
                yield parsed

    ######
    # User-facing API
    def add_filter(self, *args) -> 'BaseReader':
        """
        Limit the output to rows that match the specified criterion. Can apply multiple filters.
        Filters are applied after all parsing and transforms are complete.

        There are three ways to specify a filter:
        - `add_filter('field_name') requires that the value not be missing. Equivalent to "is not None".
        - `add_filter('field_name', value)` checks that the field has this exact value.
        - `add_filter(lambda parsed: bool)` runs a user-provided function on whatever is in this row.

        The `field_name` should match a field in whatever object the parser returns. (attribute, calc'd property, etc)
        """
        if len(args) == 1:
            spec = args[0]
            if isinstance(spec, collections.abc.Callable):  # type: ignore
                self._filters.append(spec)
            elif isinstance(spec, str):
                self._filters.append(lambda parsed: getattr(parsed, spec) is not None)
            else:
                raise exceptions.ConfigurationException('Single argument must be either a function or a field name')
        elif len(args) == 2:
            # Exact value match
            field_name, target_value = args
            self._filters.append(lambda parsed: getattr(parsed, field_name) == target_value)
        else:
            raise exceptions.ConfigurationException('Invalid filter format requested')
        return self

    def add_lookup(self, field_name: str, lookup_func: ty.Callable[[object], object]) -> 'BaseReader':
        """
        Look up / modify the value of an individual field within a row. Each lookup is a function that receives
          the parsed row and returns a single value. This can be used to clean up how values are presented
          ("chr1" vs "chr1"), and also for things that a generic parser may be ill-suited to infer from the text of the
          file. (eg, rsid lookups require knowing reference build, which is usually specified as metadata).

        Lookups are good at finding the value of a single field. Lookups can be written that have no dependence on
          zorp, eg "genelocator".

        Example: `add_lookup('rsid', lookup_func)` sets `parsed.rsid` to the result of the lookup.

        At present, zorp is written with the restrictive philosophy that each field has a well-understood universal
          meaning, consistent across all files that use that parser. Lookups are allowed to modify existing known
          fields, but not add new fields. (a side effect: the parser must support name-based field access)
        """
        if not self._parser:
            raise exceptions.ConfigurationException(
                "Lookup features require specifying a parser that supports name-based field access.")

        # Sanity check if possible, otherwise, just hope the parser returns something with named fields!
        if hasattr(self._parser, 'fields') and field_name not in self._parser.fields:  # type: ignore
            raise exceptions.ConfigurationException("The parser does not have a field by this name")

        if not isinstance(lookup_func, collections.abc.Callable):  # type: ignore
            raise exceptions.ConfigurationException(
                "Lookup must specify a function that receives variant data and returns a value")

        self._lookups.append([
            field_name,
            lookup_func
        ])
        return self

    def add_transform(self, transform_func: ty.Callable[[object], object]) -> 'BaseReader':
        """
        Transforms provide a way to change ALL of the data for a variant. They can mutate several fields at once,
          perform multiple lookups, or even return a different object than what the parser normally returns. (this last
          option is not recommended for both performance and debugging reasons)

        We generally recommend lookups for most use cases, because they don't need to depend on your parser,
          and are easier to maintain. (because they are pure functions with no side effects). Transforms represent
          a way to make very powerful arbitrary changes to all your data at once, but they also bypass type checks and
          other features designed to prevent bugs.
        """
        if not isinstance(transform_func, collections.abc.Callable):  # type: ignore
            raise exceptions.ConfigurationException(
                "Transforms must specify a function that operates on variant data")
        self._transforms.append(transform_func)
        return self

    def write(self,
              out_fn: str = None, *,
              columns: ty.Iterable[str] = None,
              delimiter: str = '\t',
              make_tabix: bool = False):
        """
        Write an output file with the specified columns. Column names must match the field names or tuple indices
        used by your parser.

        This method is useful when you want to save your filtered results for later, or normalize different input
        formats into a single predictable output format.

        FIXME: In tabix mode, out_fn != result_fn . That is... just needlessly confusing.

        TODO: In the initial version, we hardcode a preset mode of "vcf" for tabix indexing.
            There's no good parser-agnostic way to set the tabix options, so we may just have to proxy the pysam kwargs

        # TODO: This isn't a universal converter, eg the column headers are still dictated by the parser (not the user)
        """
        if make_tabix:
            delimiter = '\t'

        if out_fn is None and make_tabix:
            raise exceptions.ConfigurationException('Cannot create tabix file if writing output to a stream')

        if isinstance(self._source, str) and out_fn == self._source:
            raise exceptions.ConfigurationException('Writer cannot overwrite input file')

        if columns is None:
            try:
                columns = self._parser.fields  # type: ignore
            except AttributeError:
                raise exceptions.ConfigurationException('Must provide column names to write')

        # Special case rule: The writer renders missing data (the Python value `None`) as `.`
        def repr_missing(v):
            return '.' if v is None else v

        def write_all(handle):
            """Internal helper that allows writing to either a file, or stdout"""
            # Write headers
            try:
                handle.write('#')
                handle.write(delimiter.join(str(name) for name in columns))
                handle.write('\n')
                for row in self:
                    handle.write(delimiter.join(str(repr_missing(getattr(row, field_name)))
                                                for field_name in columns))
                    handle.write('\n')
            except BrokenPipeError:  # pragma: no cover
                # When writing to stdout, some utils (like head) may close the pipe early, at which point we end writing
                return

        # Readers can write to stdout, which lets CLI scripts (like zorp-convert) use this in a pipeline
        try:
            with open(out_fn, 'w') as f:
                write_all(f)
        except TypeError:
            write_all(sys.stdout)

        if make_tabix:
            return pysam.tabix_index(out_fn, force=True, preset='vcf')
        else:
            return out_fn

    def __iter__(self) -> ty.Iterator:
        """
        The parser instance can be treated as an iterable over raw or parsed lines (based on options)
        """
        # Reset the error list on any new iteration
        self.errors = []

        iterator = self._create_iterator()
        # Advance the iterator (if applicable) so that only data is returned (not headers)
        for n in range(self._skip_rows):
            next(iterator)
        return self._make_generator(iterator)


class IterableReader(BaseReader):
    """
    Read data from an externally provided iterable object (like piping from sys.stdin, or a list)

    This can be used as an adapter to wrap an external source in a standard interface, and is also useful for
        unit testing.
    """
    def _create_iterator(self):
        return iter(self._source)


class TextFileReader(BaseReader):
    """Implement additional helper methods unique to working with uncompressed text files"""
    def __init__(self, *args, **kwargs):
        super(TextFileReader, self).__init__(*args, **kwargs)

    def _create_iterator(self):
        """Open the file for parsing"""
        with open(self._source, 'r') as f:
            yield from f


class TabixReader(BaseReader):
    """
    A reader that can handle any gzipped text file. If a `filename.tbi` index is present in the same folder,
    it unlocks additional region-based fetch features.
    """
    def __init__(self, *args, **kwargs):
        super(TabixReader, self).__init__(*args, **kwargs)
        self._tabix = None  # type: pysam.TabixFile
        self._has_index = bool(self._source and os.path.isfile('{}.tbi'.format(self._source)))

    def _create_iterator(self):
        """
        Return an iterator over all rows of the file
        """
        with gzip.open(self._source, 'rt') as f:
            yield from f

    def fetch(self, chrom: str, start: int, end: int) -> ty.Iterable:
        """Fetch data from the file in a specific region, by wrapping tabix functionality to return parsed data."""
        if not self._has_index:
            raise FileNotFoundError("You must generate a tabix index before using region-based fetch")

        if not self._tabix:
            # Allow repeatedly fetching from the same tabix file
            self._tabix = pysam.TabixFile(self._source)

        iterator = self._tabix.fetch(chrom, start, end)
        return self._make_generator(iterator)



"""
Exceptions related to reading and parsing data
"""


class BaseZorpException(Exception):
    DEFAULT_MESSAGE = ''  # type: str

    def __init__(self, message=None, *args):
        super(BaseZorpException, self).__init__(*args)
        self.message = message or self.DEFAULT_MESSAGE

    def __str__(self):
        return str(self.message)


class ConfigurationException(BaseZorpException):
    DEFAULT_MESSAGE = 'Invalid option specified'


class SnifferException(BaseZorpException):
    DEFAULT_MESSAGE = 'Could not auto-detect file format'


class LineParseException(BaseZorpException):
    """An error occurred while parsing a single line. Capture the data that failed to parse for context."""
    DEFAULT_MESSAGE = 'Could not parse specified line'

    def __init__(self, *args, line=None):
        super(LineParseException, self).__init__(*args)
        self.line = line


class TooManyBadLinesException(BaseZorpException):
    DEFAULT_MESSAGE = 'Too many lines in the file failed to parse; stopping'

    def __init__(self, *args, error_list: list = None, **kwargs):
        super(TooManyBadLinesException, self).__init__(*args)
        self.error_list = error_list



"""
Certain features (such as lookups) require large files that are not included with this package.

This module governs the building and downloading of such assets.
"""
from filefetcher import AssetCLI, AssetManager


# For routine use, instantiate a manager. It will locate cached copies of asset files.
# FIXME: Temporary download URL; will eventually move this to an S3 bucket
manager = AssetManager('zorp', 'https://csg-assets.occsci.com/zorp/manifest.json')  # site hosts a manifest.json file


def set_recipes():
    """
    The Zorp use case assumes that all assets are built in advance, so we will only define recipes
        when explicitly asked to do so (such as in CLI mode)
    """

    from .loaders.make_rsid_lookup import MakeSnpToRsid

    B37_SAMPLE_GENES = (
        ('2', 21224301, 21266945),  # APOB
        ('19', 45409039, 45412650),  # APOE
        ('17', 41196312, 41277500),  # BRCA1
        ('13', 32889617, 32973809),  # BRCA2
        ('10', 135340300, 135352627),  # CYP2E1
        ('16', 53737875, 54148379),  # FTO
        ('15', 28356183, 28567313),  # HERC2
        ('1', 55505149, 55530526),  # PCSK9
        ('10', 114709978, 114927437),  # TCF7L2
    )

    B38_SAMPLE_GENES = (
        ('2', 21001429, 21044073),  # APOB
        ('19', 44905796, 44909395),  # APOE
        ('17', 43044295,  43125364),  # BRCA1
        ('13', 32315508, 32400268),  # BRCA2
        ('10', 133527363, 133539123),  # CYP2E1
        ('16', 53703963, 54121941),  # FTO
        ('15', 28111040, 28322179),  # HERC2
        ('1', 55039548, 55064853),  # PCSK9
        ('10', 112950247, 113167678),  # TCF7L2
    )

    manager.add_recipe(
        'snp_to_rsid',
        MakeSnpToRsid('GRCh37'),
        label='Find rsID information given chrom/pos/ref/alt',
        genome_build='GRCh37'
    )

    manager.add_recipe(
        'snp_to_rsid_test',
        MakeSnpToRsid('GRCh37', sample_regions=B37_SAMPLE_GENES),
        label='Find rsID information given chrom/pos/ref/alt',
        genome_build='GRCh37'
    )

    manager.add_recipe(
        'snp_to_rsid',
        MakeSnpToRsid('GRCh38'),
        label='Find rsID information given chrom/pos/ref/alt',
        genome_build='GRCh38'
    )

    manager.add_recipe(
        'snp_to_rsid_test',
        MakeSnpToRsid('GRCh38', sample_regions=B38_SAMPLE_GENES),
        label='Find rsID information given chrom/pos/ref/alt',
        genome_build='GRCh38'
    )


def main():
    """
    Main method. Register recipes and run the CLI.
    """
    set_recipes()

    cli = AssetCLI(manager)
    cli.run()


if __name__ == '__main__':
    main()



"""
Parsers: handle the act of reading one entity (such as line)
"""
import math
import numbers
import typing as ty

try:
    from fastnumbers import int, float
except ImportError:  # pragma: no cover
    pass

from .const import MISSING_VALUES
from . import exceptions, parser_utils as utils

# Whitelist of allowed chromosomes. It's ok to add more values, as long as we have some kind of whitelist.
#  The generic parser uses these as a safeguard, because when people slip a non-categorical value into the chrom field,
#  tabix uses all the RAM on the system and then crashes horribly.
ALLOWED_CHROMS = frozenset({
    '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20',
    '21', '22', '23', '24', '25',
    'X', 'Y', 'M', 'MT'
})


class BasicVariant:
    """
    Store GWAS results in a predictable format, with a minimal set of fields; optimize for name-based attribute access
    """
    # Slots specify the data  this holds (a performance optimization); _fields is human-curated list
    __slots__ = ('chrom', 'pos', 'ref', 'alt', 'neg_log_pvalue', 'beta', 'stderr_beta', 'alt_allele_freq', 'rsid')
    _fields = ('chrom', 'pos', 'rsid', 'ref', 'alt', 'neg_log_pvalue', 'beta', 'stderr_beta', 'alt_allele_freq')

    def __init__(self, chrom, pos, rsid, ref, alt, neg_log_pvalue, beta, stderr_beta, alt_allele_freq):
        self.chrom = chrom  # type: str
        self.pos = pos  # type: int
        self.rsid = rsid  # type: str
        self.ref = ref  # type: str
        self.alt = alt  # type: str
        self.neg_log_pvalue = neg_log_pvalue  # type: float

        # # Optional fields for future expansion
        # af: float
        self.beta = beta  # type: float
        self.stderr_beta = stderr_beta  # type: float

        self.alt_allele_freq = alt_allele_freq

    @property
    def pvalue(self) -> ty.Union[float, None]:
        if self.neg_log_pvalue is None:
            return None
        elif math.isinf(self.neg_log_pvalue):
            # This is an explicit design choice here, since we parse p=0 to infinity
            return 0
        else:
            return 10 ** -self.neg_log_pvalue

    @property
    def pval(self) -> numbers.Number:
        """A common field name alias"""
        return self.pvalue

    @property
    def maf(self) -> ty.Union[numbers.Number, None]:
        af = self.alt_allele_freq
        return min(af, 1 - af) if af is not None else None

    @property
    def marker(self) -> str:
        """Specify the marker in a string format compatible with UM LD server and other variant-specific requests"""
        ref_alt = '_{}/{}'.format(self.ref, self.alt) \
            if (self.ref and self.alt) else ''
        return '{}:{}{}'.format(self.chrom, self.pos, ref_alt)

    def to_dict(self):
        # Some tools expect the data in a mutable form (eg dicts)
        return {s: getattr(self, s, None) for s in self._fields}


def TupleLineParser(*args, container: ty.Callable = tuple, delimiter='\t', **kwargs):
    """
    Parse a line of text and return a tuple of the fields. Performs no type coercion

    This isn't recommended for everyday parsing, but it is useful internally for format detection (where we need to
        split columns of data, but aren't yet ready to clean and assign meaning to the values)
    """
    def inner(line: str):
        """Return a stateful closure that actually does the work of parsing"""
        try:
            values = line.strip().split(delimiter)
            return container(values)
        except Exception as e:
            raise exceptions.LineParseException(str(e), line=line)
    return inner


def GenericGwasLineParser(
        *args,
        delimiter: str = '\t',
        container: ty.Callable[..., BasicVariant] = BasicVariant,
        # Variant identifiers: marker OR individual
        chrom_col: int = None, chr_col: int = None,  # Legacy alias
        pos_col: int = None,
        ref_col: int = None,
        alt_col: int = None,
        marker_col: int = None,
        # Other required data
        pvalue_col: int = None, pval_col: int = None,  # Legacy alias
        # Optional fields
        rsid_col: int = None,
        beta_col: int = None,
        stderr_beta_col: int = None,

        allele_freq_col: int = None,  # As freq OR by two count fields
        allele_count_col: int = None,
        n_samples_col: int = None,
        # Other configuration options that apply to every row as constants
        is_neg_log_pvalue: bool = False, is_log_pval: bool = False,  # Legacy alias
        is_alt_effect: bool = True,  # whether effect allele is oriented towards alt
        **kwargs):
    """
    A simple parser that extracts GWAS information from a flexible file format.

    Constructor expects human-friendly column numbers (first = column 1)
    """
    def validate_config():
        """Ensures that a minimally working parser has been created"""
        # Some old gwas files may not have ref and alt (incomplete marker, or missing columns). These fields aren't
        #   strictly required, but we really really like to have them
        has_position = (_marker_col is not None) ^ all(x is not None
                                                       for x in (_chrom_col, _pos_col))
        # If we do have one allele, we must have both
        both_markers = (_ref_col is None and _alt_col is None) or \
                       (_ref_col is not None and _alt_col is not None)

        is_valid = has_position and both_markers and (_pvalue_col is not None)
        if not is_valid:
            raise exceptions.ConfigurationException('GWAS parser must specify how to find all required fields')

        if _allele_count_col is not None and _allele_freq_col is not None:
            raise exceptions.ConfigurationException('Allele count and frequency options are mutually exclusive')

        if _allele_count_col is not None and _n_samples_col is None:
            raise exceptions.ConfigurationException(
                'To calculate allele frequency from counts, you must also provide n_samples')

        return is_valid

    def inner(line):
        # Return a stateful closure that does the actual work of parsing
        try:
            fields = line.strip().split(delimiter)
            if len(fields) == 1:
                raise exceptions.LineParseException(
                    'Unable to split line into separate fields. This line may have a missing or incorrect delimiter.')

            # Fetch values
            ref = None
            alt = None
            if _marker_col is not None:
                chrom, pos, ref, alt = utils.parse_marker(fields[_marker_col])
            else:
                chrom = fields[_chrom_col]
                pos = fields[_pos_col]

            if chrom.startswith('chr'):
                chrom = chrom[3:]

            chrom = chrom.upper()

            if chrom not in ALLOWED_CHROMS:
                options = ' '.join(utils.natural_sort(ALLOWED_CHROMS))
                raise exceptions.LineParseException(
                    f"Chromosome {chrom} is not a valid option. Must be one of: '{options}'")

            # Explicit columns will override a value from the marker, by design
            if _ref_col is not None:
                ref = fields[_ref_col]

            if _alt_col is not None:
                alt = fields[_alt_col]

            pval = fields[_pvalue_col]

            # Some optional fields
            rsid = None
            beta = None
            stderr_beta = None
            alt_allele_freq = None
            allele_count = None
            n_samples = None

            if _rsid_col is not None:
                rsid = fields[_rsid_col]
                if rsid in MISSING_VALUES:
                    rsid = None
                elif not rsid.startswith('rs'):
                    rsid = 'rs' + rsid

            if _beta_col is not None:
                beta = fields[_beta_col]

            if _stderr_col is not None:
                stderr_beta = fields[_stderr_col]

            if _allele_freq_col is not None:
                alt_allele_freq = fields[_allele_freq_col]

            if _allele_count_col is not None:
                allele_count = fields[_allele_count_col]
                n_samples = fields[_n_samples_col]

            # Perform type coercion
            log_pval = utils.parse_pval_to_log(pval, is_neg_log=_is_neg_log_pvalue)

            try:
                pos = int(pos)
            except ValueError:
                # Some programs seem to write long positions using scientific notation, which int cannot handle
                try:
                    pos = int(float(pos))
                except ValueError:
                    # If we still can't parse, it's probably bad data
                    raise exceptions.LineParseException(
                        'Positions should be specified as integers. Could not parse value: {}'.format(pos))

            if beta is not None:
                beta = None if beta in MISSING_VALUES else float(beta)
            if stderr_beta is not None:
                stderr_beta = None if stderr_beta in MISSING_VALUES else float(stderr_beta)

            if _allele_freq_col or _allele_count_col:
                alt_allele_freq = utils.parse_allele_frequency(
                    freq=alt_allele_freq,
                    allele_count=allele_count,
                    n_samples=n_samples,
                    is_alt_effect=_is_alt_effect
                )

            # Some old GWAS files simply won't provide ref or alt information, and the parser will need to do without
            if ref in MISSING_VALUES:
                ref = None

            if isinstance(ref, str):
                ref = ref.upper()

            if alt in MISSING_VALUES:
                alt = None

            if isinstance(alt, str):
                alt = alt.upper()

            result = container(chrom, pos, rsid, ref, alt, log_pval, beta, stderr_beta, alt_allele_freq)
        except Exception as e:
            raise exceptions.LineParseException(str(e), line=line)
        return result

    # Convert the user-provided values to field array indices (0-based), and validate config
    # Chrom field has a legacy alias, allowing older parser configs to work.
    inner._chrom_col = _chrom_col = utils.human_to_zero(chrom_col) if chrom_col is not None else utils.human_to_zero(chr_col)  # type: ignore  # noqa
    inner._pos_col = _pos_col = utils.human_to_zero(pos_col)  # type: ignore
    inner._ref_col = _ref_col = utils.human_to_zero(ref_col)  # type: ignore
    inner._alt_col = _alt_col = utils.human_to_zero(alt_col)  # type: ignore

    inner._marker_col = _marker_col = utils.human_to_zero(marker_col)  # type: ignore

    # Support legacy alias for field name
    inner._pvalue_col = _pvalue_col = utils.human_to_zero(pvalue_col) if pvalue_col is not None else utils.human_to_zero(pval_col)  # type: ignore  # noqa

    inner._rsid_col = _rsid_col = utils.human_to_zero(rsid_col)  # type: ignore
    inner._beta_col = _beta_col = utils.human_to_zero(beta_col)  # type: ignore
    inner._stderr_col = _stderr_col = utils.human_to_zero(stderr_beta_col)  # type: ignore

    inner._allele_freq_col = _allele_freq_col = utils.human_to_zero(allele_freq_col)  # type: ignore
    inner._allele_count_col = _allele_count_col = utils.human_to_zero(allele_count_col)  # type: ignore
    inner._n_samples_col = _n_samples_col = utils.human_to_zero(n_samples_col)  # type: ignore

    # The latter option is an alias for legacy reasons
    inner._is_neg_log_pvalue = _is_neg_log_pvalue = is_neg_log_pvalue or is_log_pval  # type: ignore
    inner._is_alt_effect = _is_alt_effect = is_alt_effect  # type: ignore

    # Raise an exception if the provided options are invalid
    validate_config()

    # Provide the outside world with access to additional named attributes
    # We are slightly abusing closures, but the end result is ~10% is faster than a class-based callable
    inner.fields = container._fields  # type: ignore
    return inner



#!/usr/bin/env python
"""
Look up the RSID associated with a variant, based on a tabixed copy of dbSNP

This script is provided for internal development reference. It is not a fully-fledged lookup utility and should not
    be relied on for production use. Some filenames/ chrom versions may be hard-coded, etc.

TODO: Small changes to data pipeline have been made; verify that the `LookupRsidsTabix` still works
"""
import typing as ty

import pysam

from zorp import sniffers

from zorp import lookups
from zorp.loaders.make_rsid_lookup import make_group_iterator, make_file_iterator, make_chrom_to_contigs


class LookupRsidsTabix:
    """
    Find RSIDs that match a specified file. This is a tabix-based design and inherently assumes that each lookup
        performed will be sequential (after the previous one)
    """
    def __init__(self, source_fn: str, max_segment: int = 50000):
        # Define the data to iterate over
        self._tabix = pysam.TabixFile(source_fn)
        self._current_reader = None  # type: ty.Iterable[ty.Tuple[str, ty.Any, ty.Dict[ty.Any, ty.Any]]]

        # Store the "current" row (sometimes our dbsnp reader will get AHEAD of the sumstats reader,
        #   or be asked for the same position twice in a row, eg multiallelic variants)
        self._current_row = None  # type: ty.Tuple[str, ty.Any, ty.Dict[ty.Any, ty.Any]]

        # Optimization: actual dbSNP is a very large file, and we want to only iterate over the parts near variants
        #   in the sumstats file
        # To avoid excessive disk seeks, we'll prefetch 50kb regions at a time

        # TODO: Investigate tabix file
        #   1. Get the chromosome names from the tabix index provided by NIH (remember, columns are a bit weird)
        #   2. Use chromosome ranges to grab tabix data (requires pysam instead of gzip reader)

        # Identify where the last tabix query was run
        self._last_query_chrom = None  # In human readable format so we can quickly compare to the current row
        self._last_query_pos = 0
        self._max_segment = max_segment  # type: int

    def make_reader(self, chrom, start_pos):
        """Make a tabix reader from the data for the specified region"""
        human_chrom_to_tabix = make_chrom_to_contigs(self._tabix)
        dnsnp_chrom = human_chrom_to_tabix[chrom]
        segment = self._tabix.fetch(dnsnp_chrom, start_pos, start_pos + self._max_segment)
        all_lines = make_file_iterator(segment)

        # Update current reader position
        self._current_reader = iter(make_group_iterator(all_lines))
        self._current_row = next(self._current_reader)

        self._last_query_chrom = chrom
        self._last_query_pos = start_pos

    def _advance_current_reader(self, target_chrom, target_pos):
        # 1. Decide whether to advance the current reader (a 50kb chunk of data) or seek to a new section of the
        #   file on disk
        if self._last_query_chrom != target_chrom or target_pos > (self._last_query_pos + self._max_segment):
            self.make_reader(target_chrom, target_pos)

        # 2. Advance the reader to the desired position
        chrom, pos, _ = self._current_row

        if chrom == target_chrom and pos >= target_pos:
            # Don't advance the dbSNP reader any further- we are already at or past where we need to be
            return self._current_row

        # If the dbSNP reader is lagging the gwas reader, then we need to advance
        while chrom != target_chrom or pos < target_pos:
            try:
                self._current_row = next(self._current_reader)  # type: ignore
            except StopIteration:
                # If we hit the end of the dbSNP file, exit the loop and never let the iterator advance again
                break

            chrom, pos, _ = self._current_row

    def __call__(self, chrom: str, pos: int, ref: str, alt: str) -> ty.Union[int, None]:
        # Look up the rsid associated with a current SNP.
        self._advance_current_reader(chrom, pos)

        snp_chrom, snp_pos, ref_alt_options = self._current_row
        if snp_chrom != chrom or snp_pos != pos:
            # Ensure that the dbSNP reader is at the correct position!
            return None
        return ref_alt_options.get('{}/{}'.format(ref, alt), None)

    def __del__(self):
        self._tabix.close()


if __name__ == '__main__':
    ####
    # Local development (uses part of sumstats, with a matching section of dbsnp)
    # gwas_reader = sniffers.guess_gwas_standard(
    #     '/Users/abought/code/personal/zorp/loaders/benchmarks/data/summary_stats_1k_lines.tab'
    # )
    # rsid_finder = lookups.SnpToRsid(
    #     '/Users/abought/code/personal/zorp/loaders/benchmarks/data/dbsnp_b153_fragment_for_sumstats_benchmarks.lmdb'
    # )
    # rsid_finder = LookupRsidsTabix(
    #     '/Users/abought/code/personal/zorp/loaders/benchmarks/data/dbsnp_b153_fragment_for_sumstats_benchmarks.gz'
    # )

    ###
    # Options useful on the cluster (full dataset)
    gwas_reader = sniffers.guess_gwas_standard('/home/abought/code/zorp/loaders/benchmarks/data/summary_stats.gz')
    # rsid_finder = LookupRsidsTabix('/home/abought/dbsnp/b153/GCF_000001405.25.gz')
    rsid_finder = lookups.SnpToRsid('/home/abought/dbsnp/b153/dbSNP_grch37_b153.lmdb')

    # Perform lookups and track results
    gwas_reader.add_lookup(
        'rsid', lambda variant: rsid_finder(variant.chrom, variant.pos, variant.ref, variant.alt)  # type: ignore
    )

    gwas_reader.write('deleteme.gz')



"""
Convert SNP information to gene, rsid, or other useful annotations
"""
import struct
import typing as ty

import lmdb
import msgpack

from . import assets, exceptions


class SnpToRsid:
    """Convert SNP coordinates to RSID information"""
    def __init__(self, path_or_build: str, *, num_chroms: int = 25, test=False):
        if not path_or_build:
            raise exceptions.ConfigurationException('Must provide a path to the lookup file')

        if path_or_build == 'GRCh37' or path_or_build == 'GRCh38':
            # Certain special strings will automatically trigger opening a particular file
            record_type = 'snp_to_rsid'
            if test:
                record_type += '_test'
            path_or_build = assets.manager.locate(record_type, genome_build=path_or_build)

        self.env = lmdb.open(path_or_build, subdir=False, max_dbs=num_chroms, readonly=True)
        self.db_handles = {}  # type: dict

        self._known_chroms = None  # type: set

    def __call__(self, chrom: str, pos: int, ref: str, alt: str) -> ty.Union[int, None]:
        """
        Look up the specified SNP in the database, and handle any translation of how results are stored in this
            specific file format.

        Assumes an LMDB database file with several child databases (one per chromosome).

        Each record is an integer key (position), and a msgpack object of { refalt_str: rsid_int } entries
        This deserialization adds some overhead but also seems to reduce total file size on the (large) lmdb file
        """
        if chrom not in self.known_chroms:
            return None

        if chrom not in self.db_handles:
            self.db_handles[chrom] = self.env.open_db(bytes(chrom, 'utf8'), integerkey=True)

        db = self.db_handles[chrom]

        key = struct.pack('I', int(pos))
        with self.env.begin(buffers=True) as txn:
            res = txn.get(key, db=db)
            if res:  # If there is a match for this position, find any matching ref/alts
                res = msgpack.unpackb(res, use_list=False)
                res = res.get('{}/{}'.format(ref, alt))

            if res is not None:  # If this exact snp has a match, format it as `rsINT_VALUE` for display
                res = 'rs{}'.format(res)
        return res

    @property
    def known_chroms(self) -> set:
        if self._known_chroms is None:
            with self.env.begin() as txn:
                # In this particular data structure, all child databases are listed as keys in the default DB
                cursor = txn.cursor()
                self._known_chroms = {k.decode('utf-8') for k, v in cursor}
        return self._known_chroms

    def __del__(self):
        # Ensure the database environment is closed when the reader is no longer needed
        try:
            self.env.close()
        except Exception:
            pass



#!/usr/bin/env python3
"""
Generate an LMDB database optimized for chr:pos:ref:alt --> rsid lookups

Sample usage:
    ./make_rsid_lookups.py dbSNPvcf.gz dbSNPb153.lmbd --force


Each database file is structured as one file, multiple databases (one DB per chromosome)
Each per-chromosome DB has one key per (int)position, and each value is a hash of possible {"ref/alt":rsid} matches.
Positions and rsids are represented as integers to save space in an effort to stay below virtual memory limits.

The final database format (one database per chromosome) is:

{  pos:  { 'ref/alt1': 1 , 'ref/alt2': 2 }  (where rs1 and rs2 are represented as integers, "1" and "2")
"""
import gzip
import itertools
import os
import re
import struct
import tempfile
import typing as ty
import urllib.request

from fastnumbers import int
from filefetcher.manager import BuildTask
import lmdb
import msgpack
import pysam


RSID_CAPTURE = re.compile(r'RS=(\d+);?')

# The new dbSNP format uses refseq identifiers. Building a lookup based on human friendly chrom/pos/ref/alt
#  requires converting human identifiers to and from refseq names
VERSIONLESS_CHROMS = {
    'NC_000001': '1',
    'NC_000002': '2',
    'NC_000003': '3',
    'NC_000004': '4',
    'NC_000005': '5',
    'NC_000006': '6',
    'NC_000007': '7',
    'NC_000008': '8',
    'NC_000009': '9',
    'NC_000010': '10',
    'NC_000011': '11',
    'NC_000012': '12',
    'NC_000013': '13',
    'NC_000014': '14',
    'NC_000015': '15',
    'NC_000016': '16',
    'NC_000017': '17',
    'NC_000018': '18',
    'NC_000019': '19',
    'NC_000020': '20',
    'NC_000021': '21',
    'NC_000022': '22',
    'NC_000023': 'X',
    'NC_000024': 'Y',
    'NC_012920': 'MT',
}


def make_chrom_to_contigs(tabix_file: pysam.TabixFile) -> dict:
    """
    In order to address a tabix file by position, we need the exact contig name. dbSNP builds use a versioned
        identifier that changes across builds and versions.

    This generates a lookup for converting human friendly chromosome names to things that can be used by a
        particular dbSNP file. It automatically adjusts for small build differences and excludes non-chrom contigs
    """
    return {
        # Eg '1': 'NC_000001.10',
        VERSIONLESS_CHROMS[c.split('.', 1)[0]]: c
        for c in tabix_file.contigs if c.startswith('NC')
    }


def fetch_regions_sequentially(source_fn: str, sample_regions: ty.Tuple[str, int, int]) -> ty.Iterator[str]:
    """Instead of loading an entire gzip file, create an iterator over a set of regions"""
    source = pysam.TabixFile(source_fn)
    human_chrom_to_tabix = make_chrom_to_contigs(source)

    for region in sample_regions:
        chrom, start, end = region
        dbsnp_chrom = human_chrom_to_tabix[chrom]

        for line in source.fetch(dbsnp_chrom, start, end):
            yield line


def line_parser(row) -> ty.Tuple[str, int, str, str, int]:
    """For new dbSNP format, builds 152+"""
    fields = row.split()
    # the new dbSNP format uses refseq ids + version; convert these to human-readable chromosome names
    chrom = VERSIONLESS_CHROMS[fields[0].split('.')[0]]
    pos = int(fields[1])

    ref = fields[3]
    alt = fields[4]

    # Get the RSID from the VCF info field, in case the id column is ambiguous for some reason
    rsid = int(RSID_CAPTURE.search(fields[7]).group(1))

    return (chrom, pos, ref, alt, rsid)


def make_file_iterator(handle: ty.Iterable) -> ty.Iterator[ty.Tuple[str, int, str, str, int]]:
    """
    Parse the set of lines for some iterator (eg over contents of an open text, gz, etc file), and only give back the
        ones relevant to our chrom/pos/ref/alt use case
    """
    for row in handle:
        if row.startswith('#'):
            # Skip header rows
            continue

        if not row.startswith('NC_'):
            # After that, only parse the variants labeled "NC", not the "unplaced scaffold" items
            raise StopIteration

        yield line_parser(row)


def make_group_iterator(file_iterator) -> ty.Iterable[ty.Tuple[str, int, dict]]:
    """Create an iterator that returns all possible ref/alt : rsid pairs for that position"""

    chrom = None
    for position, position_records in itertools.groupby(file_iterator, key=lambda x: x[1]):
        # assumes that position is the only thing that will change, else use chr/pos: (x[0], x[1])):
        position_contents = {}
        # Then push final lookup for that chr:pos into the database
        for record in position_records:
            # The same line can indicate one or more ref/alts
            chrom, pos, ref, alt, rsid = record
            ref_alts = ['{}/{}'.format(ref_option, alt_option)
                        for alt_option in alt.split(',')
                        for ref_option in ref.split('.')]

            for key in ref_alts:
                position_contents[key] = rsid

        yield chrom, position, position_contents


def make_databases(env) -> dict:
    known = [
        '1', '2', '3', '4', '5', '6', '7', '8', '9', '10',
        '11', '12', '13', '14', '15', '16', '17', '18', '19', '20',
        '21', '22', 'X', 'Y', 'MT'
    ]
    return {k: env.open_db(bytes(k, "utf8"), integerkey=True) for k in known}


def main(source_fn: str, out_fn: str, n_chroms=25, sample_regions=None):
    """Perform this task in isolation for any dbsnp file"""
    if os.path.exists(out_fn):
        # LMDB would happily append / replace keys, but we want to create these files once and keep file size small
        raise Exception('The requested output file already exists.')

    # To reduce disk usage, each chromosome is stored internally as a separate database (the same chrom key info isn't
    #   stored redundantly)
    env = lmdb.open(out_fn, subdir=False, max_dbs=n_chroms, map_size=25e9)
    db_handles = make_databases(env)

    if sample_regions:
        iterator = fetch_regions_sequentially(source_fn, sample_regions)
    else:
        iterator = gzip.open(source_fn, "rt")

    with env.begin(write=True) as txn:
        all_lines = make_file_iterator(iterator)
        group_iterator = make_group_iterator(all_lines)

        for chrom, position, position_contents in group_iterator:
            # Value is not a primitive; serialize it efficiently
            key = struct.pack('I', position)
            value = msgpack.packb(position_contents, use_bin_type=True)
            txn.put(key, value, db=db_handles[chrom])


class MakeSnpToRsid(BuildTask):
    """A packaged filefetcher build task that also downloads the necessary input files"""
    def __init__(self, genome_build, sample_regions=None):
        self.genome_build = genome_build
        self.regions = sample_regions or None

    def get_assets(self):
        # Download the appropriate dbsnp file for a given genome build if does not exist
        # Uses a system temp directory rather than the build folder in case file is needed between runs

        # FIXME: The directory structure of new dbSNP is still too new to know what will change between releases, so
        #  we don't yet have a generic way to get build information without downloading the files. For
        #  now this is a hardcoded reference.
        if self.genome_build == 'GRCh37':
            source_url = 'ftp://ftp.ncbi.nih.gov/snp/redesign/latest_release/VCF/GCF_000001405.25.gz'
        elif self.genome_build == 'GRCh38':
            source_url = 'ftp://ftp.ncbi.nih.gov/snp/redesign/latest_release/VCF/GCF_000001405.38.gz'
        else:
            raise Exception('Unknown build')

        dest_fn = os.path.join(tempfile.gettempdir(), os.path.basename(source_url))

        if not os.path.exists(dest_fn):
            # Download assets to a tempfile directory
            urllib.request.urlretrieve(
                source_url,
                dest_fn
            )
            # And also Tabix index
            urllib.request.urlretrieve(
                source_url + '.tbi',
                dest_fn + '.tbi'
            )
        return dest_fn

    def build(self, manager, item_type: str, build_folder: str, **kwargs):
        # FIXME: Hardcode dbsnp build until we build a more reliable way to get this info.
        dbsnp_build = 'b153'

        source_fn = self.get_assets()
        print('Building to: ', build_folder)
        dest_fn = os.path.join(
            build_folder,
            '{}_{}_{}.lmdb'.format(item_type, self.genome_build, dbsnp_build)
        )

        main(source_fn, dest_fn, sample_regions=self.regions)

        return dest_fn, {'dbsnp_build': dbsnp_build}



"""
Auto-detect the appropriate options for parsing a GWAS file
"""

# pvalue or logpvalue
# How many header rows? (in a text file, we don't necessarily know this)
# What are the headers names?
# Get columns for chrom/pos/ref/alt
import binascii
import itertools
import typing as ty

try:
    from fastnumbers import float
except ImportError:  # pragma: no cover
    pass

from .const import MISSING_VALUES
from . import (
    const,
    exceptions,
    parsers,
    parser_utils as utils,
    readers
)


def is_numeric(val: str) -> bool:
    """Check whether an unparsed string is a numeric value"""
    if val in MISSING_VALUES:
        return True

    try:
        float(val)
    except Exception:
        return False
    else:
        return True


def is_header(row: str, *, comment_char="#", delimiter='\t') -> bool:
    """
    This assumes two basic rules: the line is not a comment, and gwas data is more likely to be numeric than headers
    """
    return row.startswith(comment_char) or all(not is_numeric(field)
                                               for field in row.split(delimiter))


def levenshtein(s1, s2):
    # CC_BY_SA https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python
    if len(s1) < len(s2):
        return levenshtein(s2, s1)

    if len(s2) == 0:
        return len(s1)

    previous_row = list(range(len(s2) + 1))
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            # j+1 instead of j since previous_row and current_row are one character longer than s2
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def find_column(column_synonyms: tuple, header_names: list, threshold: int = 2) -> ty.Union[int, None]:
    #  Find the column name that best matches
    best_score = threshold + 1
    best_match = None
    for i, header in enumerate(header_names):
        if header is None:
            # If header is empty, don't consider it for a match
            # Nulling a header provides a way to exclude something from future searching
            continue

        score = min(levenshtein(header, s) for s in column_synonyms)
        if score < best_score:
            best_score = score
            best_match = i
    return best_match


def get_pval_column(header_names: list, data_rows: ty.Iterable, overrides: dict = None) -> dict:
    """
    Return (column, is_neg_log), or None/None if config not found
    Column numbers are 1-based, for compatibility with parsers
    """
    overrides = overrides or {}

    LOGPVALUE_FIELDS = ('neg_log_pvalue', 'log_pvalue', 'log_pval', 'logpvalue')
    PVALUE_FIELDS = ('pvalue', 'p.value', 'p-value', 'pval', 'p_score', 'p', 'p_value')

    data = itertools.islice(data_rows, 100)

    def _validate_p(col: int, data: ty.Iterator, is_log: bool) -> bool:
        # All values must be parseable
        vals = [row[col] for row in data]
        cleaned_vals = [None if val in const.MISSING_VALUES else val
                        for val in vals]
        try:
            for v in cleaned_vals:
                utils.parse_pval_to_log(v, is_neg_log=is_log)
        except Exception:
            return False

        return True

    # Overrides will "win" if present
    manual_pcol = utils.human_to_zero(overrides.get('pvalue_col'))
    manual_islog = overrides.get('is_neg_log_pvalue')

    log_p_col = (manual_islog and manual_pcol) or find_column(LOGPVALUE_FIELDS, header_names)
    p_col = (not manual_islog and manual_pcol) or find_column(PVALUE_FIELDS, header_names)

    if log_p_col is not None and _validate_p(log_p_col, data, True):
        return {'pvalue_col': log_p_col + 1, 'is_neg_log_pvalue': True}
    elif p_col is not None and _validate_p(p_col, data, False):
        return {'pvalue_col': p_col + 1, 'is_neg_log_pvalue': False}

    # Could not auto-determine an appropriate pvalue column
    return {}


def get_chrom_pos_ref_alt_columns(header_names: list, data_rows: ty.Iterable, overrides: dict = None) -> dict:
    """
    Find the information required to uniquely identify the marker. Returns 1-based column indices for compatibility
        with line parsers.
    """
    overrides = overrides or {}

    # Get from either a marker, or 4 separate columns
    MARKER_FIELDS = ('snpid', 'marker', 'markerid', 'snpmarker', 'chr:position')
    CHR_FIELDS = ('chrom', 'chr', 'chromosome')
    POS_FIELDS = ('position', 'pos', 'begin', 'beg', 'bp', 'end', 'ps', 'base_pair_location')

    data = itertools.islice(data_rows, 100)

    # Order matters: consider ambiguous field names for ref before alt
    REF_FIELDS = ('A1', 'ref', 'reference', 'allele0', 'allele1')
    ALT_FIELDS = ('A2', 'alt', 'alternate', 'allele1', 'allele2')

    first_row = next(data)
    marker_col = utils.human_to_zero(overrides.get('marker_col')) or find_column(MARKER_FIELDS, header_names)

    if marker_col is not None and utils.parse_marker(first_row[marker_col], test=True):
        return {'marker_col': marker_col + 1}

    # If single columns were incomplete, attempt to auto detect 4 separate columns. All 4 must
    #  be found for this function to report a match.
    headers_marked = header_names.copy()
    to_find = [
        ['chrom_col', CHR_FIELDS, True],
        ['pos_col', POS_FIELDS, True],
        ['ref_col', REF_FIELDS, False],
        ['alt_col', ALT_FIELDS, False],
    ]
    config = {}
    for col_name, col_choices, is_required in to_find:
        col = utils.human_to_zero(overrides.get(col_name)) or \
              find_column(col_choices, headers_marked, threshold=1)  # type: ignore
        if col is None and is_required:
            return {}
        if col is not None:
            config[col_name] = col + 1
            # Once a column has been assigned, remove it from consideration for future matches
            headers_marked[col] = None

    return config


def get_effect_size_columns(header_names: list, data_rows: ty.Iterable, overrides: dict = None):
    overrides = overrides or {}

    BETA_FIELDS = ('beta', 'effect_size', 'alt_effsize', 'effect')
    STDERR_BETA_FIELDS = ('stderr_beta', 'stderr', 'sebeta', 'effect_size_sd', 'se', 'standard_error')

    data = itertools.islice(data_rows, 100)

    def _validate_numeric(col: int, data: ty.Iterator) -> bool:
        vals = [row[col] for row in data]
        cleaned_vals = [val for val in vals
                        if val not in const.MISSING_VALUES]
        try:

            for v in cleaned_vals:
                float(v)
        except Exception:
            return False
        return True

    beta_col = utils.human_to_zero(overrides.get('beta_col')) or find_column(BETA_FIELDS, header_names, threshold=0)
    stderr_col = utils.human_to_zero(overrides.get('stderr_beta_col')) or find_column(STDERR_BETA_FIELDS,
                                                                                      header_names, threshold=0)

    ret = {}
    if beta_col is not None and _validate_numeric(beta_col, data):
        ret['beta_col'] = beta_col + 1

    if stderr_col is not None and _validate_numeric(stderr_col, data):
        ret['stderr_beta_col'] = stderr_col + 1

    return ret


def get_reader(filename: ty.Union[ty.Iterable, str]) -> ty.Type[readers.BaseReader]:
    """Suggest an appropriate reader class based on source of data: iterable, tabix, or text file"""
    if not isinstance(filename, str):
        return readers.IterableReader

    with open(filename, 'rb') as test_f:
        # A known magic number for GZIP files: simple filetype detection
        is_gz = binascii.hexlify(test_f.read(2)) == b'1f8b'

    if is_gz:
        return readers.TabixReader
    else:
        return readers.TextFileReader


def get_headers(reader, comment_char: str = "#", delimiter: str = '\t', max_check=100) \
        -> ty.Tuple[int, ty.Union[str, None]]:
    """Identify the number of header rows, and the content of the one likely to contain column headers"""
    last_row = None
    for i, row in enumerate(reader):
        # TODO: Move some of this method to parser class to keep the reader domain-agnostic
        if not is_header(row, comment_char=comment_char, delimiter=delimiter):
            return i, last_row
        elif i >= max_check:
            raise exceptions.SnifferException('No headers found after limit of {} rows'.format(max_check))
        last_row = row

    raise exceptions.SnifferException('No headers found after searching entire file')


def guess_gwas_generic(filename: ty.Union[ty.Iterable, str], *,
                       skip_rows=None,
                       parser: ty.Callable[[str], object] = None,
                       parser_options: dict = None,
                       delimiter: str = '\t',
                       **kwargs) -> readers.BaseReader:
    """
    Read tab delimited rows of data and return a fully configured reader/parser
    This will try to identify the following options:
    - Type of file (text, gzip, etc)
    - Number of header rows (unless provided explicitly)
    - How to parse the file (unless a parser is explicitly provided). Because some columns vary too much to sniff,
        a "parser_overrides" option can be used to provide extra info used to create the parser.

    Supports receiving an iterable (instead of filename), primarily to support unit testing
    """
    reader_class = get_reader(filename)
    n_headers, header_text = get_headers(reader_class(filename, parser=None), delimiter=delimiter)

    parser_options = parser_options or {}
    parser_options = {k: v for k, v in parser_options.items() if v is not None}  # all kwargs must have values

    if parser and parser_options:
        raise exceptions.ConfigurationException(
            'You have specified an exact `parser` and partial `parser_options`. These options are mutually exclusive.')

    # Don't try to guess parser options if options are explicitly provided. That would be silly.
    to_skip = n_headers if skip_rows is None else skip_rows
    if parser is None:
        # FIXME: Handle case of files with no header rows
        header_names = header_text.lower().strip().lstrip('#').split(delimiter)

        # The first effort at field detection just extracts fields, with no value cleanup
        parser = parsers.TupleLineParser(delimiter=delimiter)

        # Any kwargs not specified for this function are assumed to be reader options, and passed directly in
        data_reader = reader_class(filename, skip_rows=to_skip, parser=parser, **kwargs)

        p_config = get_pval_column(header_names, data_reader, overrides=parser_options)
        if not p_config:
            raise exceptions.SnifferException('Could not find required field: pvalue')

        header_names[p_config['pvalue_col'] - 1] = None  # Remove this column from consideration for other matches
        position_config = get_chrom_pos_ref_alt_columns(header_names, data_reader, overrides=parser_options)

        if not position_config:
            raise exceptions.SnifferException('Could not find SNP identifier columns (position or marker)')

        for v in position_config.values():
            header_names[v - 1] = None  # Remove columns from consideration

        beta_config = get_effect_size_columns(header_names, data_reader, overrides=parser_options)

        # Configure a reader and parser based on the auto-detected file options, plus any explicit argument overrides
        options = {**p_config, **position_config, **(beta_config or {}), **(parser_options or {})}

        parser = parsers.GenericGwasLineParser(**options)

    return reader_class(filename, skip_rows=to_skip, parser=parser, **kwargs)


def guess_gwas_standard(filename: ty.Union[ty.Iterable, str], *,
                        parser: ty.Callable[[str], object] = None,
                        parser_options: dict = None,
                        delimiter: str = '\t',
                        **kwargs) -> readers.BaseReader:
    """
    Return a fully configured reader/parser for the harmonized GWAS format used by zorp

    This is not a tool for generic formats; it is used to locate optional columns and adjust column numbers when
        present. Because it trusts the user to only use this on standard data, some datatype validation is skipped.

    This will try to identify the following options:
    - Type of file (text, gzip, etc)
    - How to parse the file (unless a parser is explicitly provided). A "parser_overrides" option can be used to
      provide extra info for the parser (eg, container)

    Supports receiving an iterable (instead of filename), primarily to support unit testing
    """
    reader_class = get_reader(filename)
    n_headers, header_text = get_headers(reader_class(filename, parser=None), delimiter=delimiter)

    header_names = header_text.lower().strip().lstrip('#').split(delimiter)

    parser_options = parser_options or {}
    parser_options = {k: v for k, v in parser_options.items() if v is not None}  # all kwargs must have values

    if parser and parser_options:
        raise exceptions.ConfigurationException(
            'You have specified an exact `parser` and partial `parser_options`. These options are mutually exclusive.')

    default_parser_options = {'is_neg_log_pvalue': True, 'is_alt_effect': True}
    column_config = {}

    required_cols = [
        ['chrom', 'chrom_col'], ['pos', 'pos_col'],
        ['neg_log_pvalue', 'pvalue_col']
    ]
    optional_cols = [
        ['ref', 'ref_col'], ['alt', 'alt_col'],
        ['beta', 'beta_col'], ['stderr_beta', 'stderr_beta_col'],
        ['alt_allele_freq', 'allele_freq_col'],
        ['rsid', 'rsid_col']
    ]

    # if not all(name in header_names for name, _ in required_cols):
    #     raise exceptions.SnifferException('File must specify all columns required by the standard format')

    for header, out_field in required_cols:
        try:
            index = header_names.index(header)
            column_config[out_field] = index + 1
        except ValueError:
            raise exceptions.SnifferException(
                'File must specify all columns required by the standard format. Missing: {}'.format(header))

    for header, out_field in optional_cols:
        try:
            index = header_names.index(header)
            column_config[out_field] = index + 1
        except ValueError:
            pass

    options = {**column_config, **default_parser_options, **parser_options}
    parser = parsers.GenericGwasLineParser(**options)  # type: ignore
    return reader_class(filename, skip_rows=n_headers, parser=parser, **kwargs)
